/**RLFM.cpp***********************************************************************************************
 *          Copyright 2004, V. Makinen                                                                   *
 *                                                                                                       *
 * This file contains an implementation of a succinct full-text index: RLFM-index.           	      	 * 
 * For updates, see: http://www.cs.helsinki.fi/u/vmakinen/software                                       *
 * For more information, see [MN04].        	      	                                                 *                     
 *                                                                                                       *
 *                                                                                                       *   
 * This software may be used freely for any non-commercial purpose. However,                             *
 * when distributed, the original source must be clearly stated, and, when the                           *
 * source code is distributed, the copyright notice must be retained and any                             *
 * alterations in the code must be clearly marked. No warranty is given regarding                        *
 * the quality of this software.                                                                         *
 *                                                                                                       *
 * [MN04] Veli Mkinen and Gonzalo Navarro. New search algorithms and space/time tardeoffs               *    
 *        for succinct suffix arrays                                                                     *
 *        Technical report C-2004-20, Dept. CS, Univ. Helsinki, April 2004.    	      	             	 *
 *                                                                                                       *
 *********************************************************************************************************/


//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
#include "maketable.cpp"
//---------------------------------------------------------------------------

unsigned n,m; // text length, pattern length
bool print=true; 
int report=1;  // 0=existence, 1=report, 2=count, X>2 = print X characters after each pattern occurrence
unsigned bitsn=0; // bits needed to represent n
char* text; 

class TSA;
class TAlphabetRank;
char *BWT(char *text, unsigned n);
void TestAlphabetRank();
TSA *CreateSuffixArray(char *text, unsigned n);

/**********************************************************************************************************
 * Some preliminary stuff                                                                                 *
 **********************************************************************************************************/

#define mask31 0x0000001F

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))


#define W 32 //numero de bits del entero de la maquina
#define Wminusone 31  // W-1
#define WW 64 //numero de bits del entero de la maquina
#define bitsM 8 //bits para hacer la mascara para contar mas rapido
#define BW 4 //bytes que hacen una palabra

const unsigned char __popcount_tab[] =
{
0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8,
};

const unsigned char select_tab[] =
{
0,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

7,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

8,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

7,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,
};


// bits needed to represent a number between 0 and n
inline uint bits (uint n){
	uint b = 0;
	while (n) { b++; n >>= 1; }
	return b;
}

// reads bit p from e
#define bitget(e,p) ((((e)[(p)/W] >> ((p)%W))) & 1)
// sets bit p in e
#define bitset(e,p) ((e)[(p)/W] |= (1<<((p)%W)))
// cleans bit p in e
#define bitclean(e,p) ((e)[(p)/W] &= ~(1<<((p)%W)))

// numero de enteros necesarios para representar e elementos de largo n
#define enteros(e,n) ((e)*(n))/W+(((e)*(n))%W > 0)

inline unsigned GetField(unsigned *A, register  unsigned len, register unsigned index) {
register unsigned i=index*len/W, j=index*len-W*i, result;
if (j+len <= W)
result = (A[i] << (W-j-len)) >> (W-len);
else {
result = A[i] >> j;
result = result | (A[i+1] << (WW-j-len)) >> (W-len);
}
return result;
};

inline unsigned GetFieldW32(unsigned *A,register unsigned index) {
return A[index];
};

inline unsigned GetFieldW16(unsigned *A,register unsigned index) {
register unsigned i=index/2, j=(index&1)<<4, result;
result = (A[i] << (16-j)) >> (16);
return result;
};

inline unsigned GetFieldW4(unsigned *A,register unsigned index) {
register unsigned i=index/8, j=(index&0x7)<<2;
//register unsigned i=index/8, j=index*4-32*i;
return (A[i] << (28-j)) >> (28);
};

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register  unsigned x) {
   unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
};

inline void SetField32(unsigned *A, register unsigned index,register unsigned x) {
   A[index]=x;
}; 


inline uint popcount (register int x){
  return __popcount_tab[(x >>  0) & 0xff]  + __popcount_tab[(x >>  8) & 0xff]  + __popcount_tab[(x >> 16) & 0xff] + __popcount_tab[(x >> 24) & 0xff];
};

inline uint popcount16 (register int x){
  return __popcount_tab[x & 0xff]  + __popcount_tab[(x >>  8) & 0xff];
};

inline uint popcount8 (register int x){
  return __popcount_tab[x & 0xff];
};


/**********************************************************************************************************
 * Suffix array in about n log n bits                                                                     *
 **********************************************************************************************************/

class TSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned *B;    
    unsigned n;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {this->n = n; posbits = bits; pos = new unsigned[n*posbits/W+2];
       CreateBooleanArray(n);};
    void CreateBooleanArray(unsigned n)
      {B = new unsigned[n/W+2];};   
    void DeletePosArray() {delete[] pos;};
    void DeleteBooleanArray() {delete[] B;};        
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    bool GetB(unsigned index)
      {return GetField(B,1,index);};        
    ~TSA() {DeletePosArray(); DeleteBooleanArray();};
};

TSA *CreateSuffixArray(char *text, unsigned n)
{
   TSA *SA;
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while ((n<<k)>>(W-1) == 0) k++;
   bitsn = W-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n);
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
   return SA;
}


/**********************************************************************************************************
 * New RANK and SELECT implementations from Rodrigo Gonzales, Univ. Chile                                 *
 * I modified the code a bit adding some more useful functions                                            * 
 **********************************************************************************************************/

class BitRankF {
private:
	uint *data; //here is the bit-array
	bool owner;
	uint n,integers;
	uint factor,b,s; 
	uint *Rs; //superblock array
	unsigned char *Rb; //block array

	uint BuildRankSub(uint ini,uint fin); //internal use of BuildRank
	int BuildRank(); //crea indice para rank
public:
	BitRankF(uint *bitarray, uint n, bool owner);
	~BitRankF(); //destructor	 
	bool IsBitSet(uint i);
	uint rank(uint i); //Rank from 0 to n-1
	uint prev(uint start); // gives the largest index i<=start such that IsBitSet(i)=true 
        uint select(uint x); // gives the position of the x:th 1.
        uint SpaceRequirementInBits();	
};

/////////////
//Rank(B,i)// 
/////////////
//This Class use a superblock size of 256 bits
//and a block size of 32 bits also
//we use a 32 bit to represent Rs and 
//we use a 8 bit to represent Rb

BitRankF::BitRankF(uint *bitarray, uint n, bool owner) {
	data=bitarray;
	this->owner = owner;
	this->n=n;
	b=32; // b is a word
	factor=8; // 8 word in 256 bits
	s=b*factor;
	uint aux=(n+1)%W;
	if (aux != 0)
		integers = (n+1)/W+1;
	else 
		integers = (n+1)/W;
	BuildRank();
}

BitRankF::~BitRankF() {
	delete [] Rs;
	delete [] Rb;
	if (owner) delete [] data;
}

//Build the rank (blocks and superblocks)
int BitRankF::BuildRank(){
	uint num_sblock = n/s;
	uint num_block = n/b;
	Rs = new uint[num_sblock+1];//+1 we add the 0 pos
	Rb = new unsigned char[num_block+1];//+1 we add the 0 pos
	int j;
	Rs[0]=0;
	for (j=1;j<=num_sblock;j++) 
		Rs[j]=BuildRankSub((j-1)*factor,factor)+Rs[j-1];
	
	Rb[0]=0;
	for (int k=1;k<=num_block;k++) {
		j = k / factor;
		Rb[k]=BuildRankSub(j*factor,k%factor);
	}
}

uint BitRankF::BuildRankSub(uint ini,uint bloques){
	uint rank=0,aux;
	for(uint i=ini;i<ini+bloques;i++) {
		if (i < integers) {
			aux=data[i];
			rank+=popcount(aux);
		}
	}
	return rank; //return the numbers of 1's in the interval
}


//this rank ask from 0 to n-1
uint BitRankF::rank(uint i) {
	++i; // the following gives sum of 1s before i 
	return Rs[i>>8]+Rb[i>>5]
		+popcount(data[i>>5] & ((1<<(i & mask31))-1));
}

uint BitRankF::prev(uint start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine

      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
}

uint BitRankF::select(uint x) {
	// returns i such that x=rank(i) && rank(i-1)<x or n if that i not exist
	// first binary search over first level rank structure
	// then sequential search using popcount over a int
	// then sequential search using popcount over a char
	// then sequential search bit a bit
	
	//binary search over first level rank structure
	unsigned l=0, r=n/s;
	unsigned mid=(l+r)/2;      
	unsigned rankmid = Rs[mid];
	while (l<=r) {
		if (rankmid<x)
			l = mid+1;
		else
			r = mid-1;
		mid = (l+r)/2;              
		rankmid = Rs[mid];
	}    
	//sequential search using popcount over a int
	unsigned left;
	left=mid*factor;
	x-=rankmid;
        uint j=data[left];
        uint ones = popcount(j);
        while (ones < x) {
	  x-=ones;left++;
	  if (left > integers) return n;
          j = data[left];
  	  ones = popcount(j);
        }
	//sequential search using popcount over a char
	left=left*b; 
	rankmid = popcount8(j);
	if (rankmid < x) {
		j=j>>8;
		x-=rankmid;
		left+=8;
		rankmid = popcount8(j);
		if (rankmid < x) {
			j=j>>8;
			x-=rankmid;
			left+=8;
			rankmid = popcount8(j);
			if (rankmid < x) {
				j=j>>8;
				x-=rankmid;
				left+=8;
			}
		}
	}

	// then sequential search bit a bit
        while (x>0) {
		if  (j&1) x--;
		j=j>>1;
		left++;
	}
	return left-1;
}

bool BitRankF::IsBitSet(uint i) {
	return (1u << (i % W)) & data[i/W];
}

uint BitRankF::SpaceRequirementInBits() {
	return (owner?n:0)+W*(n/s)+bitsM*(n/b) + 9*W; // 9 pointers/variables in the class
}

/**********************************************************************************************************
 * FM-index implementation using alphabet ranking and RL-compression: RLFM-index                          *
 **********************************************************************************************************/

class TAlphabetRank {
// using fixed 0...255 alphabet
public: 
   int min; // contains ranks for symbols min...max
   int max;
   bool *B;
   BitRankF *bitrank;
   TAlphabetRank *left;
   TAlphabetRank *right;
public:
   TAlphabetRank(char *s, unsigned n, int minp, int maxp) {
      left = NULL;
      right = NULL;
      bitrank = NULL;
      B = new bool[n];
      min = minp;
      max = maxp;
      int mid = (min+max)/2;
      unsigned i, sum=0;
      for (i=0;i<n;i++) {   
         B[i] = ((int)s[i]+128)>mid;
         sum += B[i];
      }
      char *sfirst, *ssecond;
      if (n-sum > 0)  
         sfirst = new char[n-sum];
      if (sum > 0) 
         ssecond = new char[sum];
      unsigned j=0,k=0;
      for (i=0;i<n;i++)
         if (B[i]) ssecond[k++] = s[i];
         else sfirst[j++] = s[i];
      unsigned *Binbits = new unsigned[n/W+1];
      for (i=0;i<n;i++)
         SetField(Binbits,1,i,B[i]);
      delete [] B; 
      bitrank = new BitRankF(Binbits,n,true);
      if (j > 0) {
         if (min < max) 
            left = new TAlphabetRank(sfirst,j,min,mid); 
         delete [] sfirst;
      }
      if (k>0) { 
         if (min < max)           
            right = new TAlphabetRank(ssecond,k,mid+1,max); 
         delete [] ssecond;
      }
   };   
   inline unsigned rank(int c, int i) { // returns the number of characters c before and including position i
      TAlphabetRank *temp=this;
      if (c < temp->min || c >temp->max) return 0;
      while (temp!=NULL && i>=0) {
         if (c<=(temp->min+temp->max)/2) {
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
      } 
      return i+1;
   };   
   inline bool IsCharAtPos(int c, int i) {
      TAlphabetRank *temp=this;
      if (c < temp->min || c >temp->max) return false;
      while (temp!=NULL && i>=0) {
         if (c<=(temp->min+temp->max)/2) {
            if (temp->bitrank->IsBitSet(i)) return false;
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            if (!temp->bitrank->IsBitSet(i)) return false;         
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
      } 
      return true;
   }
   inline int charAtPos(int i) {
      TAlphabetRank *temp=this;
      while (temp->min!=temp->max) {
         if (temp->bitrank->IsBitSet(i)) {
	    i = temp->bitrank->rank(i)-1;
	    temp = temp->right;
	 } 
	 else {
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 	    
	 }	       
      }
      return temp->min;
   }
   unsigned SpaceRequirementInBits() {
      unsigned bits=0;
      if (left!=NULL) bits+=left->SpaceRequirementInBits();
      if (right!=NULL) bits+=right->SpaceRequirementInBits();
      bits += bitrank->SpaceRequirementInBits() +5*W; // 5 pointers/variables in the class
      return bits;
   };
   bool Test(char *s, unsigned n) {
      // testing that the code works correctly
      int C[256];
      int i,j;
      bool correct=true;
      for (j=0;j<256;j++)
         C[j] = 0;
      for (i=0;i<n;i++) {
         C[(int)s[i]+128]++;
         if (C[(int)s[i]+128]!=rank((int)s[i]+128,i)) {
            correct = false;
            printf("%d (%c): %d<>%d\n",i,s[i],C[(int)s[i]+128],rank((int)s[i]+128,i)); 
         }         
      }
      return correct;            
   };   
   ~TAlphabetRank() {
      if (left!=NULL) delete left;
      if (right!=NULL) delete right;
      if (bitrank!=NULL)   
         delete bitrank;
   };   

};


class THuffAlphabetRank {
// using fixed 0...255 alphabet
private: 
   BitRankF *bitrank;
   THuffAlphabetRank *left;
   THuffAlphabetRank *right;
   TCodeEntry *codetable;
   char ch;
   bool leaf;
public:
   THuffAlphabetRank(char *s, unsigned n, TCodeEntry *codetable, unsigned level) {
      left = NULL;
      right = NULL;
      bitrank = NULL;
      ch = s[0];
      this->codetable = codetable;
      bool *B = new bool[n];
      unsigned sum=0,i;
      for (i=0;i<n;i++) 
	 if (codetable[(int)s[i]+128].code & (1u << level)) {
	    B[i] = true;
	    sum++;
	 }
	 else B[i] = false;
      leaf=true; 
      for (i=1;i<n;i++)
         if (s[i] != ch)
	    leaf = false;
      if (leaf) {
         delete [] B;
	 return;
      } 
      char *sfirst, *ssecond;
      if (n-sum > 0)  
         sfirst = new char[n-sum];
      if (sum > 0) 
         ssecond = new char[sum];
      unsigned j=0,k=0;
      for (i=0;i<n;i++)
         if (B[i]) ssecond[k++] = s[i];
         else sfirst[j++] = s[i];
      unsigned *Binbits = new unsigned[n/W+1];
      for (i=0;i<n;i++)
         SetField(Binbits,1,i,B[i]); 
      delete [] B;
      bitrank = new BitRankF(Binbits,n,true);
      if (j > 0) { 
         left = new THuffAlphabetRank(sfirst,j,codetable,level+1); 
         delete [] sfirst;
      }
      if (k>0) {
         right = new THuffAlphabetRank(ssecond,k,codetable,level+1); 
         delete [] ssecond;
      }
   };   
   inline unsigned rank(int c, int i) { // returns the number of characters c before and including position i
      THuffAlphabetRank *temp=this;
      if (codetable[c].count == 0) return 0;
      unsigned level = 0;
      unsigned code = codetable[c].code;
      while (!temp->leaf) {
         if ((code & (1u<<level)) == 0) {
	    i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
	 ++level;
      } 
      return i+1;
   };   
   inline bool IsCharAtPos(int c, int i) {
      THuffAlphabetRank *temp=this;
      if (codetable[c].count == 0) return false;
      unsigned level = 0;
      unsigned code = codetable[c].code;      
      while (!temp->leaf) {
         if ((code & (1u<<level))==0) {
            if (temp->bitrank->IsBitSet(i)) return false;
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            if (!temp->bitrank->IsBitSet(i)) return false;         
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
	 ++level;
      } 
      return true;
   }
   inline int charAtPos(int i) {
      THuffAlphabetRank *temp=this;
      unsigned level = 0;     
      while (!temp->leaf) {
         if (temp->bitrank->IsBitSet(i)) {
	    i = temp->bitrank->rank(i)-1;
	    temp = temp->right;
	 } 
	 else {
	    i = i-temp->bitrank->rank(i); 
            temp = temp->left; 	    
	 }	       
      }
      return (int)temp->ch+128;
   }
   unsigned SpaceRequirementInBits() {
      unsigned bits=6*W; // 6 pointers/variables in the class
      if (left!=NULL) bits+=left->SpaceRequirementInBits();
      if (right!=NULL) bits+=right->SpaceRequirementInBits();
      if (bitrank != NULL)
         bits += bitrank->SpaceRequirementInBits();
      return bits;
   };
   bool Test(char *s, unsigned n) {
      // testing that the code works correctly
      int C[256];
      int i,j;
      bool correct=true;
      for (j=0;j<256;j++)
         C[j] = 0;
      for (i=0;i<n;i++) {
         C[(int)s[i]+128]++;
         if (C[(int)s[i]+128]!=rank((int)s[i]+128,i)) {
            correct = false;
            printf("%d (%c): %d<>%d\n",i,s[i],C[(int)s[i]+128],rank((int)s[i]+128,i)); 
         }         
      }
      return correct;            
   };   
   ~THuffAlphabetRank() {
      if (left!=NULL) delete left;
      if (right!=NULL) delete right;
      if (bitrank!=NULL)   
         delete bitrank;
   };   

};


char *BWT(char *text, unsigned n, unsigned samplerate, BitRankF **sampled, unsigned **suffixes, unsigned **positions) {
   TSA *SA = CreateSuffixArray(text,n);
   char *s = new char[n];
   unsigned *sampledpositions = new unsigned[n/W+1];
   (*suffixes) = new unsigned[n/samplerate+2];
   (*positions) = new unsigned[n/samplerate+2];
   unsigned i,j=0;
   for (i=0;i<n/W+1;i++)
      sampledpositions[i]=0u;
   for (i=0;i<n;i++)
      if (SA->GetPos(i) % samplerate == 0) {
         SetField(sampledpositions,1,i,1);
	 (*suffixes)[j++] = SA->GetPos(i);
	 (*positions)[SA->GetPos(i)/samplerate] = i;
      }
      else if (SA->GetPos(i)==n-1)
         (*positions)[(n-1)/samplerate+1] = i;          
   (*sampled) = new BitRankF(sampledpositions,n,true);	 
   for (i=0;i<n;i++)
      if (SA->GetPos(i)==0) s[i] = text[n-1];
      else s[i] = text[SA->GetPos(i)-1];
   delete SA;
   return s;      
}

class TFMindex {
public:
    unsigned n;
    unsigned samplerate;
    unsigned C[256];
    THuffAlphabetRank *alphabetrank;
    BitRankF *sampled; 
    unsigned *suffixes;
    unsigned *positions;
public:
    TFMindex(char *text, unsigned n, unsigned samplerate) {
       printf("Text size is %d B\n",n);        
       this->n = n;
       this->samplerate = samplerate;
       char *bwt = BWT(text,n,samplerate,&sampled,&suffixes,&positions);
       // caller can delete text now
       unsigned i,min, max;
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<n;++i)
          C[(int)bwt[i]+128]++;
       for (i=0;i<256;i++)
          if (C[i]>0) {min = i; break;}          
       for (i=255;i>=min;--i)
          if (C[i]>0) {max = i; break;}                    
       unsigned prev=C[0], temp;
       C[0]=0;
       for (i=1;i<256;i++) {          
          temp = C[i];
          C[i]=C[i-1]+prev;
          prev = temp;
       }
       TCodeEntry *codetable = makecodetable(bwt,n);
       alphabetrank = new THuffAlphabetRank(bwt,n,codetable,0);   
       //if (alphabetrank->Test(bwt,n)) printf("alphabetrank ok\n");    
       delete [] bwt;
       // to avoid running out of unsigned, the sizes are computed in specific order (large/small)*small
       // |class TFMindex| +256*|TCodeEntry|+|C[]|+|suffixes[]+positions[]|+...       
       printf("FMindex takes %d B\n",6*W/8+256*3*W/8+256*W/8+(2*n/(samplerate*8))*W+sampled->SpaceRequirementInBits()/8+alphabetrank->SpaceRequirementInBits()/8);
       fflush(stdout);       
    };   
    int Search(char *pattern, unsigned m) {
       // use the FM-search replacing function Occ(c,1,i) with alphabetrank->rank(c,i)
       int c = (int)pattern[m-1]+128; unsigned i=m-1;
       int sp = C[c];
       int ep = C[c+1]-1;
       while (sp<=ep && i>=1) {
          c = (int)pattern[--i]+128;
          sp = C[c]+alphabetrank->rank(c,sp-1);
          ep = C[c]+alphabetrank->rank(c,ep)-1;
       }
       if (sp<=ep) {
          if (print) printf("%d matches found for \"%s\"\n",ep-sp+1, pattern);
	  if (report==1 || report>2) {
	     i=sp;
	     unsigned j,k,dist,skip;
	     char *context;
	     if (report > 2) {
	        context = new char[m+2*report];
		context[m+2*report] = '\0';
	     } 
	     while (i<=ep) {
	        j=i,dist=0;
	        while (!sampled->IsBitSet(j)) {
		   c = alphabetrank->charAtPos(j);
	           j = C[c]+alphabetrank->rank(c,j)-1; // LF-mapping  
		   ++dist;
		}
		if (dist>=samplerate || suffixes[sampled->rank(j)-1] % samplerate != 0) printf("error\n");
		if (print)
		   printf("Match at %d",suffixes[sampled->rank(j)-1]+dist);
		if (report > 2) {
		   k = suffixes[sampled->rank(j)-1]+dist+m+report;
		   if ((k/samplerate+1)*samplerate>n-1) {
		      skip = n-1-k; //can be negative
		      k = n-1;		      		 
		   }
		   else
		      skip = samplerate - k % samplerate; 
		   j = positions[k/samplerate+1];
	           dist = 0;
	           for (;dist<skip+m+2*report;dist++) {
		      c = alphabetrank->charAtPos(j);;
		      j = C[c]+alphabetrank->rank(c,j)-1; // LF-mapping
		      if (dist>= skip)
		         context[m+2*report-dist+skip-1] = (int)c-128;
		   }
		   if (print)
		      printf(": %s\n",context);
		   	   		   
		}
		else if (print)
		   printf("\n");
		++i;		
	     }
 	     if (report>2)
               delete [] context;
	  }	  
          return ep-sp+1;
       } 
       else {
          if (print) printf("Pattern not found\n");
          return 0;
       }
    };   
    ~TFMindex() {
       delete alphabetrank;       
       delete sampled;
       delete suffixes;
       delete positions;
    };
};

class TRLFMindex {
public:
    unsigned n;
    unsigned len;
    unsigned C[256];
    unsigned samplerate;
    THuffAlphabetRank *alphabetrank;
    BitRankF *runs_bwt;
    BitRankF *runs_sa;
    BitRankF *sampled;
    unsigned *suffixes;
    unsigned *positions;
public:
    TRLFMindex(char *text, unsigned n, unsigned samplerate) {
       printf("Text size is %d B\n",n);
       char *bwt = BWT(text,n,samplerate,&sampled,&suffixes,&positions);
       // caller can delete text
       // remove repeats from bwt and encode runs in a bit-vector, e.g.:
       // aaabbbbaaaacccc ==> abac, 100100010001000
       // represent the same blocks in suffix array, e.g.:
       // ==> aaa aaaa bbbb cccc ==> 100100010001000
       unsigned i, min, max, j, k;
       this->n = n;
       this->samplerate = samplerate;
       len = 0;
       for (i=0;i<256;i++)
         C[i]=0;
       for (i=0;i<n;++i)
          C[(int)bwt[i]+128]++;
       bool **CC = new bool*[256];
       for (i=0;i<256;i++)
          if (C[i]>0) 
             CC[i] = new bool[C[i]];
      	  else 
	     CC[i] = NULL;
       
       char *bwts;
       bool *B = new bool[n+1];      
       B[0] = 1; ++len;
       for (i=1;i<n;i++)
          if (bwt[i]!=bwt[i-1]) { B[i] = 1; ++len;}
	  else B[i] = 0;
       B[n] = 1;
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<n;i++)
          CC[(int)bwt[i]+128][C[(int)bwt[i]+128]++] = B[i];
       bool *BB = new bool[n+1];	 	  
       k = 0;
       for (i=0;i<256;i++)
      	  for (j=0;j<C[i];j++)
	     BB[k++] = CC[i][j]; 
       BB[n] = 1;
       for (i=0;i<256;i++)
          if (CC[i]!= NULL) delete CC[i]; 
       delete CC;
       bwts = new char[len+1];
       j = 0;
       for (i=0;i<n;i++) 
          if (B[i]) bwts[j++]=bwt[i];
       delete bwt;	  
       unsigned *Binbits = new unsigned[n/W+1];
       for (i=0;i<n;i++)
          SetField(Binbits,1,i,B[i]);
       delete B; 
       runs_bwt = new BitRankF(Binbits,n,true);
       Binbits = new unsigned[n/W+1]; 
       for (i=0;i<n;i++)
          SetField(Binbits,1,i,BB[i]);
       delete BB;
       runs_sa = new BitRankF(Binbits,n,true);
       // recompute C
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<len;++i)
          C[(int)bwts[i]+128]++;
       for (i=0;i<256;i++)
          if (C[i]>0) {min = i; break;}          
       for (i=255;i>=min;--i)
          if (C[i]>0) {max = i; break;}                    
       unsigned prev=C[0], temp;
       C[0]=0;
       for (i=1;i<256;i++) {          
          temp = C[i];
          C[i]=C[i-1]+prev;
          prev = temp;
       }
       TCodeEntry *codetable = makecodetable(bwts,len);
       alphabetrank = new THuffAlphabetRank(bwts,len,codetable,0);         
       delete [] bwts;    
       // to avoid running out of unsigned, the sizes are computed in specific order (large/small)*small
       // |class TRLFMindex| +256*|TCodeEntry|+|C[]|+|suffixes[]+positions[]|+...
       printf("RLFMindex takes %d B\n",9*W/8+256*3*W/8+256*W/8+(2*n/(samplerate*8))*W+ sampled->SpaceRequirementInBits()/8 + alphabetrank->SpaceRequirementInBits()/8 + 
          runs_bwt->SpaceRequirementInBits()/8 + runs_sa->SpaceRequirementInBits()/8);
       fflush(stdout);       
    };   
    int Search(char *pattern, unsigned m) {
       // use the FM-search replacing function Occ(c,1,i) with alphabetrank->rank(c,i)
       int c = (int)pattern[m-1]+128; unsigned i=m-1;
       int rank;
       int sp = runs_sa->select(C[c]+1);
       int ep = runs_sa->select(C[c+1]+1)-1;
       while (sp<=ep && i>=1) {
          c = (int)pattern[--i]+128;
          rank = runs_bwt->rank(sp)-1;
          if (alphabetrank->IsCharAtPos(c,rank)) // if sp is inside the run of c
	     sp = sp-runs_bwt->prev(sp);
	  else 
	     sp = 0;
          sp += runs_sa->select(C[c]+alphabetrank->rank(c,rank-1)+1);
          rank = runs_bwt->rank(ep)-1;
          if (alphabetrank->IsCharAtPos(c,rank)) // if ep is inside the run of c
	     ep = ep-runs_bwt->prev(ep);
	  else 
	     ep = -1;
          ep += runs_sa->select(C[c]+alphabetrank->rank(c,rank-1)+1);
       }
       if (sp<=ep) {
          if (print) printf("%d matches found for \"%s\"\n",ep-sp+1, pattern);
	  if (report==1 || report > 2) {
	     i=sp;
	     unsigned j,k,dist;
	     int skip;
	     char *context;
	     if (report > 2) {
	        context = new char[m+2*report];
		for (j=0;j<m+2*report; j++)
   		   context[j] = '\0';
	     } 	      
	     while (i<=ep) {
	        j=i,dist=0;
	        while (!sampled->IsBitSet(j)) { // LF-mapping
		   rank = runs_bwt->rank(j)-1;
		   c = alphabetrank->charAtPos(rank);
		   j = runs_sa->select(C[c]+alphabetrank->rank(c,rank))+j-runs_bwt->prev(j);
		   ++dist;
		}
		if (dist>=samplerate || suffixes[sampled->rank(j)-1] % samplerate != 0) printf("error\n");		
		if (print)
		   printf("Match at %d",suffixes[sampled->rank(j)-1]+dist);
		if (report > 2) {
		   k = suffixes[sampled->rank(j)-1]+dist+m+report;
		   if ((k/samplerate+1)*samplerate>n-1) {
		      skip = n-1-k; //can be negative
		      k = n-1;		      		 
		   }
		   else
		      skip = samplerate - k % samplerate; 
		   j = positions[k/samplerate+1];
                   dist = 0;
	           for (;dist<skip+m+2*report;dist++) {
		      rank = runs_bwt->rank(j)-1;
		      c = alphabetrank->charAtPos(rank);
		      j = runs_sa->select(C[c]+alphabetrank->rank(c,rank))+j-runs_bwt->prev(j);
		      if (dist >= skip)
		         context[m+2*report-dist+skip-1] = (int)c-128;
		   }
		   if (print)
		      printf(": %s\n",context);		   	   		   
		}
		else if (print)
		   printf("\n");		   
		++i;		
	     }
 	     if (report>2)
               delete context;   
          }	  
          return ep-sp+1;
       }       
       else {
          if (print) printf("Pattern not found\n");
          return 0;
       }
    };   
    ~TRLFMindex() {
       delete alphabetrank;      
       delete runs_bwt;
       delete runs_sa;        
       delete sampled;
       delete [] suffixes;
       delete [] positions;
    };
};



